/* Public domain, like most of the rest of ksh */

/*
 * area-based allocation built on malloc/free
 */

#include "sh.h"

struct link {
	struct link *prev;
	struct link *next;
};

Area *
ainit(Area *ap)
{
	ap->freelist = NULL;
	return ap;
}

void
afreeall(Area *ap)
{
	struct link *l, *l2;

	for (l = ap->freelist; l != NULL; l = l2) {
		l2 = l->next;
		free(l);
	}
	ap->freelist = NULL;
}

#define L2P(l)	( (void *)(((char *)(l)) + sizeof(struct link)) )
#define P2L(p)	( (struct link *)(((char *)(p)) - sizeof(struct link)) )

static void
nomem(void)
{
	errno = ENOMEM;
	internal_errorf(1, "unable to allocate memory");
}

size_t
size_add(size_t a, size_t b)
{
	/* This should just check against SIZE_MAX, but the code has pervasive
	 * use of pointer subtractions and int variables for sizes. */
	if (a > (size_t)INT_MAX || b > (size_t)INT_MAX - a)
		nomem();
	return a + b;
}

void *
alloc(size_t size, Area *ap)
{
	struct link *l;

	l = malloc(size_add(sizeof(struct link), size));
	if (l == NULL)
		nomem();
	l->next = ap->freelist;
	l->prev = NULL;
	if (ap->freelist)
		ap->freelist->prev = l;
	ap->freelist = l;

	return L2P(l);
}

/*
 * This is sqrt(SIZE_MAX+1), as s1*s2 <= SIZE_MAX
 * if both s1 < MUL_NO_OVERFLOW and s2 < MUL_NO_OVERFLOW
 */
#define MUL_NO_OVERFLOW	(1UL << (sizeof(size_t) * 4))

void *
areallocarray(void *ptr, size_t nmemb, size_t size, Area *ap)
{
	if ((nmemb >= MUL_NO_OVERFLOW || size >= MUL_NO_OVERFLOW)
			&& nmemb > 0
			&& SIZE_MAX / nmemb < size)
		nomem();

	return aresize(ptr, nmemb * size, ap);
}

void *
aresize(void *ptr, size_t size, Area *ap)
{
	struct link *l, *l2, *lprev, *lnext;

	if (ptr == NULL)
		return alloc(size, ap);

	l = P2L(ptr);
	lprev = l->prev;
	lnext = l->next;

	l2 = realloc(l, size_add(sizeof(struct link), size));
	if (l2 == NULL)
		nomem();
	if (lprev)
		lprev->next = l2;
	else
		ap->freelist = l2;
	if (lnext)
		lnext->prev = l2;

	return L2P(l2);
}

void
afree(void *ptr, Area *ap)
{
	struct link *l, *l2;

	if (!ptr)
		return;

	l = P2L(ptr);

	for (l2 = ap->freelist; l2 != l; l2 = l2->next)
		if (!l2)
			/* likely memory corruption so don't try to recover */
			abort();

	if (l->prev)
		l->prev->next = l->next;
	else
		ap->freelist = l->next;
	if (l->next)
		l->next->prev = l->prev;

	free(l);
}
