/* xtermsize: configure TTY driver window size by querying XTerm-alikes
 * J. Welsh, 2017
 *
 * Useful for serial consoles
 */

#include <sys/types.h>
#include <termios.h>
#include <fcntl.h>
#include <errno.h>

ssize_t read(int fd, void *buf, size_t count);
ssize_t write(int fd, const void *buf, size_t count);
size_t strlen(const char *s);
void *memcpy(void *dest, const void *src, size_t n);

int write_all(int fd, const void *buf, size_t count) {
	ssize_t n;
	while (count) {
		if ((n = write(fd, buf, count)) == -1) return -1;
		buf += n;
		count -= n;
	}
	return 0;
}

typedef struct {
	char *data;
	size_t alloc;
	size_t len;
} buffer;

void buf_init(buffer *b, char *data, size_t alloc) {
	b->data = data;
	b->alloc = alloc;
	b->len = 0;
}

void buf_clear(buffer *b) {
	b->len = 0;
}

int buf_addch(buffer *b, char c) {
	if (b->len == b->alloc) return -1;
	b->data[b->len++] = c;
	return 0;
}

int buf_addcstr(buffer *b, const char *s) {
	size_t n, new_len, pos;
	n = strlen(s);
	new_len = b->len + n;
	if (new_len > b->alloc || new_len < b->len) return -1;
	memcpy(b->data + b->len, s, n);
	b->len = new_len;
	return 0;
}

int buf_write(buffer *b, int fd) {
	return write_all(fd, b->data, b->len);
}

int is_digit(char c) {
	return c >= '0' && c <= '9';
}

int digit_val(char c) {
	return c - '0';
}

typedef enum { START, ESC, CSI, CSI8, HEIGHT0, HEIGHT, WIDTH0, WIDTH, DONE } state_t;

int main() {
	state_t state = START;
	unsigned h, w;
	char c;
	int ret = 0, tty;
	static struct termios tio, tio_saved;
	static char _b[1024];
	buffer b;
	buf_init(&b, _b, sizeof(_b));

	if ((tty = open("/dev/tty", O_RDWR)) == -1) return 1;
	tcgetattr(tty, &tio_saved);
	tio.c_oflag = tio_saved.c_oflag;
	tio.c_cflag = tio_saved.c_cflag;
	tio.c_cc[VTIME] = 5; /* read timeout in tenths of a second */
	cfsetispeed(&tio, cfgetispeed(&tio_saved));
	cfsetospeed(&tio, cfgetospeed(&tio_saved));
	tcsetattr(tty, TCSANOW, &tio);

	buf_addcstr(&b, "\033[18t");
	if (buf_write(&b, tty)) goto perror;
	buf_clear(&b);
	while (state != DONE) {
		switch (read(tty, &c, 1)) {
		case 0: goto error;
		case -1: goto perror;
		}
		switch (state) {
		case START: if (c == '\033') { state = ESC; continue; } break;
		case ESC:   if (c == '[') { state = CSI; continue; } break;
		case CSI:   if (c == '8') { state = CSI8; continue; } break;
		case CSI8:  if (c == ';') { state = HEIGHT0; continue; } break;
		case HEIGHT0:
			if (is_digit(c)) {
				if (buf_addcstr(&b, "LINES=")) break;
				if (buf_addch(&b, c)) break;
				h = digit_val(c);
				state = HEIGHT; continue;
			}
			break;
		case HEIGHT:
			if (c == ';') { state = WIDTH0; continue; }
			if (is_digit(c)) {
				if (buf_addch(&b, c)) break;
				h = 10*h + digit_val(c);
				continue;
			}
			break;
		case WIDTH0:
			if (is_digit(c)) {
				if (buf_addcstr(&b, "\nCOLUMNS=")) break;
				if (buf_addch(&b, c)) break;
				w = digit_val(c);
				state = WIDTH; continue;
			}
			break;
		case WIDTH:
			if (c == 't') { state = DONE; continue; }
			if (is_digit(c)) {
				if (buf_addch(&b, c)) break;
				w = 10*w + digit_val(c);
				continue;
			}
			break;
		}
		goto error;
	}
	if (buf_addch(&b, '\n')) goto perror;
	if (buf_write(&b, 1)) goto perror;
	goto cleanup;
perror:
	ret = -errno;
	goto cleanup;
error:
	ret = 1;
cleanup:
	tcsetattr(tty, TCSANOW, &tio_saved);
	return ret;
}
